namespace BetDriver.StrategyNewMarkets
{
    using System;
    using System.Collections.Generic;
    using System.ComponentModel;
    using System.Linq;

    using BetDriver.Api;
    using BetDriver.Api.Common;
    using BetDriver.Api.Localization;
    using BetDriver.Api.Rules;
    using BetDriver.Api.Strategies;
    using BetDriver.Api.Properties;

    using Xceed.Wpf.Toolkit.PropertyGrid.Attributes;

	[LocalizedDisplayName(typeof(Resources),"RSC_Stavkinanovblkhrblnkakh")]
    public class SatretegyNewMarkets : ForwardSearchAutoStrategy
    {
        public SatretegyNewMarkets()
			: base(true)
		{
			RuleSet = new RuleSet { new NewMarketsRule() };
            ProcessPeriod = TimeSpan.FromSeconds(1);
			Description = Resources.RSC_Strategiyazhdetpoyavleniya;
		}

		/// <summary>
		/// Проверить возможно ли удалить рынок.
		/// </summary>
		/// <param name="market">Рынок.</param>
		/// <returns>True, если рынок может быть удален, иначе false.</returns>
		protected override bool OnCanRemove(Market market)
		{
            return RuleSet.OfType<NewMarketsRule>().First().TryRemoveMarket(market, false);
		}

		/// <summary>
		/// Удалить рынок из обработки.
		/// </summary>
		/// <param name="market">Рынок.</param>
		protected override void OnRemoveMarket(Market market)
		{
            RuleSet.OfType<NewMarketsRule>().First().TryRemoveMarket(market, true);
		}

        /// <summary>
        /// Запустить стратегию.
        /// </summary>
        protected override void OnStart()
        {

            var rule = RuleSet.OfType<NewMarketsRule>().First();
            rule.StrategyStartTime = DateTime.Now;
            rule.CountOfPrices = 0;
            var tmpPrice = rule.MinPrice;
            while (tmpPrice <= rule.MaxPrice)
            {
                rule.CountOfPrices++;
                tmpPrice = tmpPrice.NextPrice();
            }

            base.OnStart();
        }

    }

	[LocalizedDisplayName(typeof(Resources),"RSC_Stavkinanovblkhrblnkakh")]
	[LocalizedCategory(typeof(Resources),"RSC_2Obrabotkarblnkov")]
	public class NewMarketsRule : Rule
	{
		/// <summary>
		/// Класс Ставка для данной стратегии.
		/// </summary>
		private class NewMarketsBet : Bet
		{
		}

		/// <summary>
		/// Класс RunnerInfo для хранения вспомогательной информации по исходам
		/// </summary>
		private sealed class RunnerInfo
		{
			public RunnerInfo(Runner runner)
			{
				if (runner == null)
					throw new ArgumentNullException("runner");

				Runner = runner;
				OurBets = new Dictionary<NewMarketsBet, bool>();

			}

			public Runner Runner { get; private set; }

			public Dictionary<NewMarketsBet, bool> OurBets { get; private set; }

		}

		private readonly Dictionary<Runner, RunnerInfo> _infos = new Dictionary<Runner, RunnerInfo>();

		private readonly object _sync = new object();
		private readonly HashSet<Market> _processedMarkets = new HashSet<Market>();

        public int CountOfPrices { get; set; }

		/// <summary>
		/// Запрашивать лестницу.
		/// </summary>
		public override bool RequestLadder
		{
			get { return true; }
		}

		/// <summary>
		/// Запрашивать сматченные суммы.
		/// </summary>
		public override bool RequestTradedAmounts
		{
			get { return true; }
		}

		/// <summary>
		/// Запрашивать счет
		/// </summary>
		public override bool RequestSoccerScores
		{
			get { return false; }
		}

		#region Свойства правила

		public DateTime StrategyStartTime { get; set; }

		private decimal _maxPrice;
		private decimal _minPrice;
		private BetAmount _amountInfo;
		private PriceTypes _betTypes;
		private string _runnerNames;
		private decimal _maxAmountBefore;
        private int _runnerNumber;

		private DateTime _lastCleanOldMarkets;

		/// <summary>
		/// Наименования исходов (через запятую)
		/// </summary>
		[LocalizedDisplayName(typeof(Resources),"RSC_Naimenovaniyaiskhodovchere")]
		[LocalizedCategory(typeof(Resources),"RSC_3Filtrbldlyastavok")]
		[LocalizedDescription(typeof(Resources),"RSC_Esliukazanblnaimenovaniyai")]
		[Parameter]
		[PropertyOrder(120)]
		public string RunnerNames
		{
			get { return _runnerNames; }
			set
			{
				_runnerNames = value;
				RaiseSettingsChanged();
			}
		}

        /// <summary>
        /// Номер исхода
        /// </summary>
        [LocalizedDisplayName(typeof(Resources), "RSC_RunnerNumber")]
        [LocalizedCategory(typeof(Resources), "RSC_3Filtrbldlyastavok")]
        [Parameter]
        [PropertyOrder(125)]
        public int RunnerNumber
        {
            get { return _runnerNumber; }
            set
            {
                _runnerNumber = value;
                RaiseSettingsChanged();
            }
        }

		/// <summary>
		/// Максимальная сумма впереди
		/// </summary>
		[LocalizedDisplayName(typeof(Resources),"RSC_Maksimalnayasummavperedi")]
		[LocalizedCategory(typeof(Resources),"RSC_3Filtrbldlyastavok")]
		[LocalizedDescription(typeof(Resources),"RSC_Poslerazmeshcheniyastavokv")]
		[Parameter]
		[PropertyOrder(130)]
		public decimal MaxAmountBefore
		{
			get { return _maxAmountBefore; }
			set
			{
				_maxAmountBefore = value;
				RaiseSettingsChanged();
			}
		}

		/// <summary>
		/// Тип ставок.
		/// </summary>
		[LocalizedDisplayName(typeof(Resources),"RSC_Tipstavok")]
		[LocalizedCategory(typeof(Resources),"RSC_4Stavki")]
		[PropertyOrder(100)]
		[Parameter]
		public PriceTypes BetTypes
		{
			get
			{
				return _betTypes;
			}
			set
			{
				_betTypes = value;
				RaiseSettingsChanged();
			}
		}

		/// <summary>
		/// Сумма ставки.
		/// </summary>
		[LocalizedDisplayName(typeof(Resources),"RSC_Summastavok")]
		[LocalizedCategory(typeof(Resources),"RSC_4Stavki")]
		[PropertyOrder(110)]
		[Parameter]
		public BetAmount AmountInfo
		{
			get { return _amountInfo; }
			set
			{
				if (_amountInfo != null)
					_amountInfo.PropertyChanged -= OnPropertyChanged;

				_amountInfo = value;

				if (_amountInfo != null)
					_amountInfo.PropertyChanged += OnPropertyChanged;

				RaiseSettingsChanged();
			}
		}

        /// <summary>
        /// Коэффициент ставок от
        /// </summary>
        [LocalizedDisplayName(typeof(Resources), "RSC_Koeffitsientstavok")]
        [LocalizedCategory(typeof(Resources), "RSC_4Stavki")]
        [Parameter]
        [PropertyOrder(120)]
        public decimal MinPrice
        {
            get { return _minPrice; }
            set
            {
                _minPrice = value;
                RaiseSettingsChanged();
            }
        }

        /// <summary>
        /// Коэффициент ставок до
        /// </summary>
        [LocalizedDisplayName(typeof(Resources), "RSC_Maksimalnblykoeffitsientst")]
        [LocalizedCategory(typeof(Resources), "RSC_4Stavki")]
        [Parameter]
        [PropertyOrder(130)]
        public decimal MaxPrice
        {
            get { return _maxPrice; }
            set
            {
                _maxPrice = value;
                RaiseSettingsChanged();
            }
        }

		/// <summary>
		/// Оставлять в In-Play
		/// </summary>
		private bool _keepInPlay = true;
		[LocalizedDisplayName(typeof(Resources),"RSC_OstavlyatvInPlay")]
		[LocalizedCategory(typeof(Resources),"RSC_4Stavki")]
		[PropertyOrder(140)]
		[Parameter]
		public bool KeepInPlay
		{
			get
			{
				return _keepInPlay;
			}
			set
			{
				_keepInPlay = value;
				RaiseSettingsChanged();
			}
		}

        private TimeSpan _refreshTime;

        /// <summary>
        /// Время обновления рынков
        /// </summary>
        [LocalizedDisplayName(typeof(Resources), "RSC_Vremyaobnovleniyarblnkov")]
        [Statistic]
        public TimeSpan RefreshTime
        {
            get { return _refreshTime; }
            set
            {
                _refreshTime = value;
                RaisePropertyChanged("RefreshTime");
            }
        }

        private int _marketsCount;

        /// <summary>
        /// Количество рынков в выборке
        /// </summary>
        [LocalizedDisplayName(typeof(Resources), "RSC_Kolichestvorblnkovvvblbork")]
        [Statistic]
        public int MarketsCount
        {
            get { return _marketsCount; }
            set
            {
                _marketsCount = value;
                RaisePropertyChanged("MarketsCount");
            }
        }

		#endregion

		public NewMarketsRule()
		{
			IsEnabled = true;
			MaxPrice = 1.01m;
			MinPrice = 1.01m;
			AmountInfo = new BetAmount
			{
				Amount = 4,
				Type = AmountTypes.BetAmount
			};
			BetTypes = PriceTypes.Lay;
			MaxAmountBefore = 0;
			RunnerNames = "";
            RunnerNumber = 0;
		}

		/// <summary>
		/// Удалить рынок из обработки.
		/// </summary>
		/// <param name="market">Рынок.</param>
		/// <returns>True, если рынок может быть удален, иначе false.</returns>
		public bool TryRemoveMarket(Market market, bool remove)
		{
			if (_processedMarkets.Contains(market))
				return false;

			foreach (var runner in market.Runners)
			{
				var info = _infos.TryGetValue(runner);

				if (info == null)
					continue;

				if (info.OurBets.Any())
					return false;
			}

			if (remove)
				_processedMarkets.Add(market);

			return true;
		}

		/// <summary>
		/// Обработчик, который вызывается при изменении рынка.
		/// </summary>
		/// <param name="market">Рынок.</param>
		protected override void OnMarketChanged(Market market, ApiTypes apiType)
		{

            // покажем в статистике периодичность обновления рынков
            RefreshTime = DateTime.Now - market.LastRefreshed;
            // покажем в статистике общее количество рынков
            MarketsCount = _processedMarkets.Count;

            market.LastRefreshed = DateTime.Now;

			return;
		}

		/// <summary>
		/// метод нужен для отключения получения данных по рынку
		/// </summary>
		/// <param name="market">Рынок.</param>
		protected override bool OnNeedUpdate(Market market)
		{
            if (!_processedMarkets.Contains(market))
                return true;

            return DateTime.Now - market.LastRefreshed >= TimeSpan.FromHours(1);
		}

		/// <summary>
		/// Метод вызывается при сбросе состояния правила.
		/// </summary>
		protected override void OnReseted()
		{
			_processedMarkets.Clear();
		}

		/// <summary>
		/// Обработчик, который определяет можно ли удалить рынок из обработки.
		/// </summary>
		/// <param name="market">Рынок.</param>
		/// <returns>True, если рынок можно удалить. Иначе False.</returns>
		protected override bool OnCanRemove(Market market)
		{
			lock (_sync)
			{

                if (market.IsInplay)
                {
                    _processedMarkets.Add(market);
                    return true;
                }

                var infos = market
                        .GetActiveRunners()
                        .Select(t => _infos.SafeGet(t, r => new RunnerInfo(r)))
                        .ToArray();
                 
                if (!_processedMarkets.Contains(market))
				{
					// делаем ставки
					TryPlaceBets(market);

					// убедимся, что выставили ставки на все исходы
					bool allBetsPlaced = true;

					foreach (var info in infos)
					{
						var mRunner = info.Runner;

						if (RunnerNames.Trim().Length > 0)
						{
							if (!RunnerNames.ToLower().Contains(mRunner.Title.ToLower()))
								continue;
						}

                        if (RunnerNumber > 0)
                        {
                            if (RunnerNumber > market.Runners.Count)
                                continue;
                            if (market.Runners[RunnerNumber - 1] != info.Runner)
                                continue;
                        }

                        var ourBetsCount = info.OurBets.Count(b => b.Key.State == BetStates.Active || b.Key.State == BetStates.Done);

						if (ourBetsCount < CountOfPrices)
							allBetsPlaced = false;

					}
					if (!allBetsPlaced)
						return false;
					// и вот теперь уже можно идти дальше и добавлять рынок в обработанные
				}

                if (!_processedMarkets.Contains(market))
                {
                    foreach (var info in infos)
                    {
                        var ourBets = info.OurBets.Where(b => b.Key.State == BetStates.Active);
                        if (ourBets == null)
                            continue;
                        foreach (var ourBet in ourBets)
                        {
                            if (ourBet.Key.AmountBefore == null)
                                ourBet.Key.AmountBefore = 0;
                            this.LogInfo(Resources.RSC_Rblnok0Iskhod1Summavperedi, market.FullTitle, info.Runner, ourBet.Key.AmountBefore);
                            if (ourBet.Key.AmountBefore > MaxAmountBefore)
                                BetsApi.CancelBet(ourBet.Key);
                        }
                    }
                }

                _processedMarkets.Add(market);

				if (DateTime.Now - _lastCleanOldMarkets > TimeSpan.FromMinutes(5))
				{
					// не чаще чем каждые 5 минут будем чистить коллекцию _processedMarkets, удаляя в ней рынки, которые уже точно закончились и больше не попадут в результаты поиска
					var startedMarkets = _processedMarkets.Where(m => m.StartTime.AddHours(4) < DateTime.UtcNow).ToArray();
					foreach (var tmpMarket in startedMarkets)
					{
						// удалим из коллекции _infos все данные по рынку из _processedMarkets
						foreach (Runner mRunner in tmpMarket.Runners)
						{
							_infos.RemoveWhere(r => r.Key == mRunner);
						}
						_processedMarkets.Remove(tmpMarket);
					}
					_lastCleanOldMarkets = DateTime.Now;
				}

				return market.StartTime < DateTime.UtcNow;

			}
		}

		/// <summary>
		/// Процедура выставления первоначальных ставок согласно выставленным параметрам
		/// </summary>
		/// <param name="mMarket">Рынок</param>
		/// <returns>Количество размещенных ставок без учета ограницений по транзакциям</returns>
		private void TryPlaceBets(Market mMarket)
		{

			var infos = mMarket
				.GetActiveRunners()
				.Select(t => _infos.SafeGet(t, r => new RunnerInfo(r)))
				.ToArray();

			foreach (var info in infos)
			{
				var mRunner = info.Runner;

				if (RunnerNames.Trim().Length > 0)
				{
					if (!RunnerNames.ToLower().Contains(mRunner.Title.ToLower()))
						continue;
				}

                if (RunnerNumber > 0)
                {
                    if (RunnerNumber > mMarket.Runners.Count)
                        continue;
                    if (mMarket.Runners[RunnerNumber - 1] != info.Runner)
                        continue;
                }

                var tmpPrice = _minPrice;
                while (tmpPrice <= MaxPrice)
                {
                    // далее если есть текущая ставка, то пропускаем исход
                    var ourBet = (from b in info.OurBets
                                  where (b.Key.State == BetStates.Active || b.Key.State == BetStates.Done || b.Key.State == BetStates.None || b.Key.State == BetStates.Updating)
                                  && b.Key.Price == tmpPrice
                                  select b.Key).FirstOrDefault();

                    if (ourBet == null)
                        PlaceBet(info, BetTypes, tmpPrice, getAmount(AmountInfo.Type, AmountInfo.Amount, tmpPrice));
                    
                    tmpPrice = tmpPrice.NextPrice();
                }
			}
		}

		/// <summary>
		/// Функция возвращает сумму ставки в зависимости от типа, заданной суммы и коэффициента
		/// </summary>
		/// <param name="type">Тип суммы</param>
		/// <param name="amount">Исходная сумма</param>
		/// <param name="price">Коэффициент</param>
		/// <returns>Сумма ставки</returns>
		private decimal getAmount(AmountTypes type, decimal amount, decimal price)
		{
			if (price <= 1)
				return 0;

			switch (type)
			{
				case AmountTypes.Payout:
					return Math.Round(amount / price, 2);

				case AmountTypes.Liability:
					return Math.Round(amount / (price - 1), 2);

				default:
					return Math.Round(amount, 2);
			}
		}

		/// <summary>
		/// Функция размещает ставку на биржу
		/// </summary>
		/// <param name="info">Исход</param>
		/// <param name="price">Коэффициент</param>
		/// <param name="amount">Сумма</param>
		/// <param name="parentBet">Родительская ставка</param>
		/// <returns>True - если ставка успешна размещена, False - если нет</returns>
		private void PlaceBet(RunnerInfo info, PriceTypes type, decimal price, decimal amount)
		{
			price = price.RoundPrice();

			amount = Math.Round(amount, 2);

			var newBet = new NewMarketsBet
			{
				Market = info.Runner.Market,
				Runner = info.Runner,
				Type = type,
				Price = price,
				UAmount = amount,
			};

			newBet.KeepAtInPlay = KeepInPlay;

            var amountOnPrice = info.Runner.GetAmount(price);
            if (amountOnPrice == null)
                amountOnPrice = 0;
            if (amountOnPrice <= MaxAmountBefore)
            {
                // выставлять ставки есть смысл только если на кэфе висит сумма меньше заданной максимальной
                if (!BetsApi.PlaceBet(newBet))
                    return;
            }
            else
            {
                this.LogInfo("Market: {0} amountOnPrice ({1}) > MaxAmountBefore {2}", info.Runner.Market, amountOnPrice, MaxAmountBefore);
                // помечаем ставку как Done, чтобы на этот исход и кэф больше ставки не делались и не пытались
                newBet.State = BetStates.Done;
            }

			info.OurBets.Add(newBet, false);

		}

		#region Настройки

		private void OnPropertyChanged(object sender, PropertyChangedEventArgs e)
		{
			RaiseSettingsChanged();
		}

		/// <summary>
		/// Загрузить настройки.
		/// </summary>
		/// <param name="settings">Настройки.</param>
		public override void Load(Settings settings)
		{
			MaxPrice = settings.GetValue<decimal>("MaxPrice", 1.01m);
			MinPrice = settings.GetValue<decimal>("MinPrice", 1.01m);
			AmountInfo.Amount = settings.GetValue("AmountInfoAmount", AmountInfo.Amount);
			AmountInfo.Type = settings.GetValue("AmountInfoType", AmountInfo.Type);
            RunnerNames = settings.GetValue<string>("RunnerNames", "");
            RunnerNumber = settings.GetValue<int>("RunnerNumber", 0);
			BetTypes = settings.GetValue<PriceTypes>("BetTypes", PriceTypes.Back);
			MaxAmountBefore = settings.GetValue<decimal>("MaxAmountBefore", 100);
            KeepInPlay = settings.GetValue<bool>("KeepInPlay", true);

			base.Load(settings);
		}

		/// <summary>
		/// Сохранить настройки.
		/// </summary>
		/// <param name="settings">Настройки.</param>
		public override void Save(Settings settings)
		{
			settings.SetValue("MaxPrice", MaxPrice);
			settings.SetValue("MinPrice", MinPrice);
			settings.SetValue("AmountInfoAmount", AmountInfo.Amount);
			settings.SetValue("AmountInfoType", AmountInfo.Type);
            settings.SetValue("RunnerNames", RunnerNames);
            settings.SetValue("RunnerNumber", RunnerNumber);
			settings.SetValue("BetTypes", BetTypes);
			settings.SetValue("MaxAmountBefore", MaxAmountBefore);
            settings.SetValue("KeepInPlay", KeepInPlay);

			base.Save(settings);
		}

		#endregion

		public override string ToString()
		{
			return string.Empty;
		}
	}
}
